/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/

#include <libc.h>
#include <string.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <stdlib.h> 

#include "OO7.hxx"
#include "GenParams.h"
#include "VarParams.h"
#include "baidlist.h"


//////////////////////////////////////////////////////////////////
//
// Global Variables, etc., for Benchmark Database Generation
//
//////////////////////////////////////////////////////////////////

struct rusage startUsage, endUsage;
struct timeval startWallTime;
struct timeval endWallTime;
struct timezone ignoreTimeZone;


extern void SetParams(char* configFileName);

extern double ComputeWallClockTime(struct timeval *startWallTime, 
	           struct timeval *endWallTime);

extern double ComputeUserTime(struct rusage *startUsage, 
	           struct rusage *endUsage);

extern double ComputeSystemTime(struct rusage *startUsage, 
	           struct rusage *endUsage);

int	nextAtomicId = 1;
int	nextCompositeId = 1;
int	nextComplexAssemblyId = 1;
int	nextBaseAssemblyId = 1;
int	nextModuleId = 1;
int	debugMode = TRUE;

char    *types[NumTypes] = {
		"type000", "type001", "type002", "type003", "type004",
		"type005", "type006", "type007", "type008", "type009"
	};

// initializers for various type extent pointers

Ref<DicoAtomicPartId> AtomicPart::extent_id ;
Ref<DicoAtomicPartDate> AtomicPart::extent_date ;
Ref<DicoDocument> Document::extent ;
Ref<DicoCompositePart> CompositePart::extent ;
Ref<DicoBaseAssembly> BaseAssembly::extent ;
Ref<DicoModule> Module::extent ;

#include "dictionary.hxx"

Ref<Cluster> DocCluster ;
char* DocCluName = "DocCluName" ;

// data structures to keep track of which composite parts each base
// assemblies uses.  This is done so that we can generate the base
// assemblies first and then the composite parts.  Why bother?
// In particular, if you did the composite parts first and the base
// assemblies, there would be no need for these data structures.
// This second approach worked fine for the small and medium databases
// but not for the large as the program seeked all over the disk.

BAIdList* private_cp;
BAIdList* shared_cp;

// usage:   GenDB <configFileName> [-d]

main(int argc, char** argv)
{
	if (argc < 2) 
	{
	      fprintf(stderr, "Usage: %s <configFileName> [-d]\n", argv[0]);
	      exit(1);
	}
	// See if debug mode is desired
	if ((argc == 3) && (strcmp(argv[2], "-d") == 0)) debugMode = TRUE;
	else  debugMode = FALSE;
	
	SetParams(argv[1]);

	shared_cp = new BAIdList[TotalCompParts+1];
	private_cp = new BAIdList[TotalCompParts+1];

        // get wall clock time
        gettimeofday(&startWallTime, &ignoreTimeZone);

	
	// get starting usage values.
	getrusage(RUSAGE_SELF, &startUsage);


	//////////////////////////////////////////////////////////////////
	//
	// Replace and Open the Database
	//
	//////////////////////////////////////////////////////////////////

	char *dbname = getenv("DBNAME");
	char *dbserver = getenv("DBSERVER");
	if (! dbname) {
	  dbname = "oo7" ;
	}
	yoo_connect(dbname, dbserver) ;
	
	//////////////////////////////////////////////////////////////////
	//
	// Initialize Yooda
	//
	//////////////////////////////////////////////////////////////////

	AtomicPart::extent_id = new DicoAtomicPartId ;
	yoo_setname(AtomicPart::extent_id, "AtomicPart::extent_id") ;
	AtomicPart::extent_date = new DicoAtomicPartDate ;
	yoo_setname(AtomicPart::extent_date, "AtomicPart::extent_date") ;
	Document::extent = new DicoDocument ;
	yoo_setname(Document::extent, "Document::extent") ;
	CompositePart::extent = new DicoCompositePart ;
	yoo_setname(CompositePart::extent, "CompositePart::extent") ;
	BaseAssembly::extent = new DicoBaseAssembly ;
	yoo_setname(BaseAssembly::extent, "BaseAssembly::extent") ;
	Module::extent = new DicoModule ;
	yoo_setname(Module::extent, "Module::extent") ;

	new ADM_Clu(DocCluName,YoodaDictionary->defVol) ;
	DocCluster = YoodaDictionary->dictClu.getFirst(DocCluName)->clu ;   

	//////////////////////////////////////////////////////////////////
	//
	// Do Benchmark Data Generation
	//
	//////////////////////////////////////////////////////////////////

        // set random number seed.
        srandom(1);


	// First generate the desired number of modules, each in its own
	// segment
	while (nextModuleId <= TotalModules) 
	{
	    printf("Generating module %d.\n", nextModuleId);
	    int id = nextModuleId++;
	    Module* module = new Module (id);
	}

	// Generate the desired number of composite parts in a design
	// library segment, with each part being in its own cluster

        int id;
	int clusterSize;
        CompositePart* comp;

	clusterSize = int (1.1 * (NumAtomicPerComp * (sizeof(AtomicPart) +
			          (NumConnPerAtomic * sizeof(Connection)))));

	Ref<Cluster> compClu = new Cluster(2) ;
	while (nextCompositeId <= TotalCompParts) 
	{
	    id = nextCompositeId++;

            if (debugMode) {
                 printf(">>> creating composite part %d itself\n", id);
            }

	    comp = new(compClu) CompositePart (id);

            if (debugMode) {
                 printf(">>> done creating composite part %d\n", id);
            }

	    if ((nextCompositeId % 50) == 0) 
	    {

		printf("commit xact\n");
//		yoo_commit() ;

		printf("created %d composite parts \n",nextCompositeId);

		printf("start a new transaction\n");

	    }
        }

	// Print out some useful information about what happened

	printf("=== DONE CREATING DATABASE, TOTALS WERE ===\n");
	printf("    # atomic parts\t\t%d\n", nextAtomicId-1);
	printf("    # composite parts\t\t%d\n", nextCompositeId-1);
	printf("    # complex assemblies\t%d\n", nextComplexAssemblyId-1);
	printf("    # base assemblies\t\t%d\n", nextBaseAssemblyId-1);
	printf("    # modules\t\t\t%d\n", nextModuleId-1);

	//////////////////////////////////////////////////////////////////
	//
	// Shutdown ObjectStore
	//
	//////////////////////////////////////////////////////////////////


	// Commit the transaction
	yoo_commit() ;
	

        // compute and report wall clock time
        gettimeofday(&endWallTime, &ignoreTimeZone);
        printf("Wall-Clock time to generate database: %f seconds.\n", 
	             ComputeWallClockTime(&startWallTime, &endWallTime));


	getrusage(RUSAGE_SELF, &endUsage);
	fprintf(stdout, "(CPU: %f seconds.)\n", 
	                ComputeUserTime(&startUsage, &endUsage) +
			ComputeSystemTime(&startUsage, &endUsage));


//      // create 60,000 1000 byte dummy objects
//	char* purgeVar;
//	purgeVar = getenv("PURGE");
//	if (purgeVar != NULL) {
//	    if (atoi(purgeVar) == 1) {
//        	printf("creating 60000 dummy 1,000 byte objects\n");
//		CreateDummy (60000);
//	    }
//	    else printf("not creating dummy purge collection\n");
//	}
//	else printf("not creating dummy purge collection\n");

	// Close the Database
	yoo_disconnect() ;

	// Exit
	exit(0);
}


