/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


/*****************************************************************************
******************************************************************************
**                                                                            
**        Y     Y   OOOO    OOOO   DDDDD     AA
**         Y   Y   O    O  O    O  D    D   A  A
**          Y Y    O    O  O    O  D    D  A    A 
**           Y     O    O  O    O  D    D  AAAAAA
**           Y     O    O  O    O  D    D  A    A
**           Y      OOOO    OOOO   DDDDD   A    A
**
******************************************************************************
******************************************************************************/
/*****************************************************************************
*
*  MODULE :             Cluster management
*  SCOPE :              Cluster management
*  FILE :               cluster.hxx
*  AUTHORS :            E. Abecassis
*  CREATION DATE :      August 1, 1993
*  COPYRIGHTS :        (C) E. Abecassis, 1993
*  HISTORY :
*
******************************************************************************/
/*=
Module Description
This module implements manipulation of the cluster manager. A cluster is a set 
of pages. It is used to group pages which have the same lock mode and to group 
objects in the same set of pages. This increases the clustering of objects. 
The cluster is a persistant object which is referenced by its pages. Each 
transaction of a workspace inserts exclusive locked pages in a structure and 
share locked pages which have more than a define free space. Exclusive locked 
pages are choose first. When an allocation is asked for less than one page, we 
look in this structure. If we don't found a free block, we insert a new page 
in the allocator. When an allocation is asked for more than one page, the 
request is transmitted to the VirtualDBServer. In a BlocAllocator, the
structure is a array of stack. Each entry of the array is for a specific block
size
=*/

#ifdef __GNUG__
#pragma interface
#endif
#ifndef __cluster_HXX
#define __cluster_HXX

/* IMPORTS *******************************************************************/

#ifndef __class_HXX
#include "class.hxx"
#endif

#ifndef __obj_HXX
#include "obj.hxx"
#endif

#ifndef __hem_HXX
#include "hem.hxx"
#endif

#ifndef __linkedlist_HXX
#include "linkedlist.hxx"
#endif

/*****************************************************************************/

class BlocAllocator {

  private :
  
  LinkedList _freeList[SLOT_IN_PG];// Linked list of all referenced pages
  OID _cluster ;                   // oid of the associated cluster
  
  public :

  BlocAllocator(OID&) ; 
  void insert(PageHdlMem*);        // insert a page which contains free blocks
  ObjectX* alloc(int ss) ;         // Try to allocate a bloc
  OID& ident() ;
} ; 


typedef HEM<256, 8, BlocAllocator, OID> ClusterSet ;
typedef IterHEM<256, 8, BlocAllocator, OID> IterClusterSet ;


class Cluster : public ObjectX {

  private :
  
  static ClusterSet AllocSet ;     // HEM set of the blocallocator
  int _volid ;
  
  public :
 
  Cluster(int vol) : _volid(vol) {} ; 
  void* operator new(size_t);         // Create a cluster in the Admin cluster
  void* operator new(size_t,Cluster*);// Create a new cluster in the DB  
  ObjectX* allocObject(int sizeSlot); // Allocation of an object in the cluster
  void insert(PageHdlMem* pd) ;       // Insert page in the cluster allocator 
  static void reset() ;               // Call delete on every used clusters  
  int getVol() ;                      // Return volid of the cluster
  void setVol(int v) ;                // Set the volume where it creates obj
} ;


inline BlocAllocator::BlocAllocator(OID& oid) : _cluster(oid) {
}

inline OID& BlocAllocator::ident() {
  return _cluster ;
}

inline void Cluster::insert(PageHdlMem* pd) {
  if (AllocSet[this->getOid()] == NULL) {
    AllocSet += new BlocAllocator(this->getOid()) ;
  }
  AllocSet[this->getOid()]->insert(pd) ;
}

inline void* Cluster::operator new(size_t s, Cluster* clu) {
  return ObjectX::operator new(s, clu, UNTYPED) ;
}

inline int Cluster::getVol() {
  return _volid ;
}

inline void Cluster::setVol(int v) {
  _volid = v ;
}

#endif
