/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


/*****************************************************************************
*
*  MODULE :             Socket Management
*  SCOPE :              Implements RPC++ through Socket
*  FILE :               sockrpc.hxx
*  AUTHORS :            E. Abecassis
*  CREATION DATE :      February 1, 1994
*  COPYRIGHTS :        (C) Apic Systemes, 1994
*  HISTORY :
*
******************************************************************************/

#ifdef __GNUG__
#pragma interface
#endif

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>

#include <netdb.h>

#include <generic.h>
#define proxy_SvcClient SvcClient
#define proxy(x) proxy_##x

#include "conversion.hxx"

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE 1
#endif

#ifndef sockrpc_HXX
#define sockrpc_HXX

#define SOCKBUFSIZE 8192*2
#define MAXSRV 64 


typedef enum { svcflush } SvcCtrl ;

// Class String to exchange 'char*' on the net
class SvcString {

  public :

  // Pointer to the string
  char* ref ;

  // Flag to mark allocated string which have to be deleted 
  int init ;

  // Empty string constructor
  SvcString() : ref(NULL), init(FALSE) {}

  // Copy constructor which duplicate the string
  SvcString(const SvcString& s) { if (s.ref != NULL) ref = strdup(s.ref) ; }

  // Copy constructor which take a pointer on an already allocated string
  SvcString(char* s) { init = TRUE; ref = s ; }

  // Delete the string if it has been allocated 
  ~SvcString() { if ((init==FALSE)&&(ref!=NULL)) free(ref) ; }

  // Assign by duplicate the string
  SvcString& operator=(char* s) { 
    if ((init==FALSE)&&(ref!=NULL)) { free(ref) ; }
    if (s != NULL) {
      ref = strdup(s) ;
    } else {
      ref = s ;
    }    
    return *this ;
  }

  // Return a pointer on char*
  operator char*() { return ref ; }

  // Allocate space for a string
  char* alloc(int s) {
    if ((init==FALSE)&&(ref!=NULL)) free(ref) ;
    ref = (char*)malloc(s+1); return ref ;
  }

  char& operator[](int i) { return ref[i] ; }
} ;


class SvcSocket {

  protected :

  // I/O stream for communication
  int        _fd ;

  // Asynchronous flag
  int       _async ;
  
  // Local and remote address
  u_int     _myport ;
  u_int     _toport ;
  u_int     _myaddr ;
  u_int     _toaddr ;

  // Service name
  char*      _svcname ;

  // Socket Family (AF_UNIX / AF_INET)
  int        _family ;

  // Socket name if the socket is of AF_UNIX family
  // Filter name if the socket is of AF_INET family
  char*      _sockn ;
  
  // error state : TRUE / FALSE value
  int        _error ;
  
  // Buffer management 
  char*      _buffIn ;
  char*      _pcurrIn ;
  char*      _ptopIn ;
  char*      _buffOut ;
  char*      _pcurrOut ;
  char*      _ptopOut ;

  // internal read
  int        read(int fd, char* buff, int size) ;

  // internal read
  int        write(int fd, char* buff, int size) ;

  // Function to call when connection is broken
  void       (*_brk)(SvcSocket*) ;

  // Initialise endpoint for the socket
  int        createSocket() ;

  // Reset input/output buffers
  void       initbuf() ;

  // fill the send buffer with buf
  int        sendbuf(char* buf, int len) ;

  // fill buf with the receive buffer
  int        recvbuf(char* buf, int len) ;

  // flush socket buffer
  int        overflow() ;

  // fill socket buffer
  int        underflow() ;

  // Call when connection is broken
  virtual void lostConnection() = 0 ;

  public :

  SvcSocket(char* svc, int family) ;
  
  virtual ~SvcSocket() ;

  virtual int isclient() = 0 ;
  virtual int isserver() = 0 ;
  
  // set socket asynchronous
  void async() { _async = 1 ; } 
  
  // return current error on the socket
  int error() { int res = _error; _error = FALSE; return res ; }
 
  // return true if the receive buffer is empty   
  int        inEmpty() { return (_pcurrIn - _ptopIn) == 0 ; }

  // return true if the send buffer is empty
  int        outEmpty() { return (_pcurrOut - _ptopOut) == 0 ; }

  // return the service name
  char*      service() { return _svcname ; }

  // return the file descriptor of the endpoint
  int        fd() { return _fd ; }

  // return connected local port
  u_int     myport() { return _myport ; }

  // return connected remote port
  u_int     toport() { return _toport ; }

  // return hostname
  char*      host(char*) ;

  // return remote hostname
  char*      tohost(char*) ;

  // Set function handler to call after disconnection  
  void       linkBreak(void (*fp)(SvcSocket*)) { _brk=fp ; }

  // return true if the Socket is connected
  int        isConnected() { return _fd != -1 ; }
  
  // return itself
  SvcSocket& socket() { return *this ; }

  // close the endpoint
  virtual int close() ;

  // send opaque buffer of len bytes
  int        send(char* buf, int len) ;

  // receive opaque buffer of len bytes
  int        recv(char* buf, int len) ;

  SvcSocket& operator << (SvcCtrl) {
    this->overflow() ; return *this ;
  } 
  SvcSocket& operator << (char data) {
    send((char*)&data, sizeof(char)) ; return *this ;
  }
  SvcSocket& operator << (u_char data) {
    send((char*)&data, sizeof(u_char)) ; return *this ;
  }
  SvcSocket& operator << (short data) {
    data = htons(data) ;
    send((char*)&data, sizeof(short)) ; return *this ;
  }
  SvcSocket& operator << (u_short data) {
    data = htons(data) ;
    send((char*)&data, sizeof(u_short)) ; return *this ;
  }
  SvcSocket& operator << (int data) {
    data = htonl(data) ;
    send((char*)&data, sizeof(int)) ; return *this ;
  }
  SvcSocket& operator << (u_int data) {
    data = htonl(data) ;
    send((char*)&data, sizeof(u_int)) ; return *this ;
  }
  SvcSocket& operator << (float data) {
    data = htonf(data) ;
    send((char*)&data, sizeof(float)) ; return *this ;
  }
  SvcSocket& operator << (double data) {
    data = htond(data) ;
    send((char*)&data, sizeof(double)) ; return *this ;
  }
  SvcSocket& operator << (char* data) {
    if (data != NULL) {
      int size = strlen(data) ;
      *this << size ;
      send((char*)data, size) ;
    } else {
      int size = -1 ;
      *this << size ;
    }
    return *this ;
  }
  SvcSocket& operator >> (SvcCtrl) {
    this->underflow() ; return *this ;
  }
  SvcSocket& operator >> (char &data) {
    recv((char*)&data, sizeof(char)); return *this ;
  }
  SvcSocket& operator >> (u_char &data) {
    recv((char*)&data, sizeof(u_char)); return *this ;
  }
  SvcSocket& operator >> (short &data) {
    recv((char*)&data, sizeof(short)); data = ntohs(data); return *this ;
  }
  SvcSocket& operator >> (u_short &data) {
    recv((char*)&data, sizeof(u_short)); data = ntohs(data); return *this ;
  }
  SvcSocket& operator >> (int &data) {
    recv((char*)&data, sizeof(int)); data = ntohl(data); return *this ;
  }
  SvcSocket& operator >> (u_int &data) {
    recv((char*)&data, sizeof(u_int)) ; data = ntohl(data); return *this ;
  }
  SvcSocket& operator >> (float &data) {
    recv((char*)&data, sizeof(float)) ; data = ntohf(data); return *this ;
  }
  SvcSocket& operator >> (double &data) {
    char* buff = (char*)&data ;
    int sz = sizeof(double) ;
    recv(buff, sz) ;
    data = ntohd(data);
    return *this ;
  }
  SvcSocket& operator >> (char*& data) {
    int size ;
    *this >> size ;
    if (size == -1) {
      data = NULL ;
    } else {
      data = new char[size+1] ;
      if (size > 0) recv((char*)data, size) ;
      data[size] = 0 ;
    }
    return *this ;
  }
  SvcSocket& operator >> (SvcString &data) {
    int size ;
    *this >> size ;
    if (size != -1) {
      if (data.init == 0) {
	data.alloc(size) ;
	if (size > 0) recv(data, size) ;
      } else {
	if (size > 0) recv(data, size) ;
      }
      data[size] = 0 ;
    }
    return *this ;
  }
} ;


#ifndef list_HXX
#include "list.hxx"
#endif


class SvcClient : public SvcSocket {
  
  protected :
  
  // Internal use to define service number 

  // Import inet server 	
  int importInet() ;

  // Import unix server
  int importUnix() ;
  
  public :  

  SvcClient(char* svcname, int family=AF_INET) : SvcSocket(svcname, family) {};

  int isclient() { return 1 ; }
  int isserver() { return 0 ; } 
  
  // Reimport socket server in a connectionless mode
  int reimport() ;
  
  // Import services at address defined by hostname and port number  
  int import(char* host, u_int port) ;

  // Import services defined by service name and filter
  // host and port are asked to the connection server
  int import(char* filter) ;

  // Connectionless import 
  int importCless(char* host, u_int port) ;
  int importCless(char* filter) ;

  // Call when connection is broken. Close the endpoint and call break handler
  void lostConnection() ;
} ;


class SvcServer : public SvcSocket {

  private :

  // Connection channel
  int _listenFd ;

  // TRUE if it is a connection mode export
  int _daemonConnected ;
  
  void (*services[MAXSRV])(SvcSocket&) ;

  public :

  // Internal use to define service number
  static SvcServer* _svcBase ;

  SvcServer(char* svcname, int family = AF_INET) :
  _listenFd(-1), _daemonConnected(0), SvcSocket(svcname, family)
  { 
    _svcBase = this ; 
  };

  int isserver() { return 1 ; }
  int isclient() { return 0 ; }

  int  listenFd() { return _listenFd ; }

  int  export(char* sockn) ;

  int  export(u_int port = 0) ;

  int  accept() ;

  int  close() ;

  virtual void execRpc() ;

  void lostConnection() ;
  
  void reqRegister(void (*fp)(SvcSocket&), int procId) {
    services[procId] = fp ;
  }
} ;


int SvcInit() ;


#endif
